@@ -1,15 +1,11 @@
-
-
-author_info = (
-    ('Dmitry Orlov', 'me@mosquito.su'),
-)
+author_info = (("Dmitry Orlov", "me@mosquito.su"),)
 
 package_info = "WSRPC is the RPC over WebSocket for aiohttp"
 package_license = "Apache Software License"
 
-team_email = 'me@mosquito.su'
+team_email = "me@mosquito.su"
 
-version_info = (2, 6, 0)
+version_info = (4, 0, 1)
 
 
 __author__ = ", ".join("{} <{}>".format(*info) for info in author_info)
@@ -17,11 +13,11 @@ __version__ = ".".join(map(str, version_info))
 
 
 __all__ = (
-    'author_info',
-    'package_info',
-    'package_license',
-    'team_email',
-    'version_info',
-    '__author__',
-    '__version__',
+    "author_info",
+    "package_info",
+    "package_license",
+    "team_email",
+    "version_info",
+    "__author__",
+    "__version__",
 )
