@@ -1,31 +1,43 @@
-from typing import Union
+import json
+import logging
+from asyncio import Lock
+from typing import Optional, Union
 
 import aiohttp
-import logging
-
 from yarl import URL
 
-from .tools import Lazy, dumps
-from .common import WSRPCBase, awaitable
+from .common import WSRPCBase
+from .tools import Lazy, awaitable
+
 
 log = logging.getLogger(__name__)
+SocketType = Optional[aiohttp.ClientWebSocketResponse]
 
 
 class WSRPCClient(WSRPCBase):
     """ WSRPC Client class """
 
-    def __init__(self, endpoint: Union[URL, str], loop=None, timeout=None,
-                 session: aiohttp.ClientSession = None, **kwargs):
+    def __init__(
+        self,
+        endpoint: Union[URL, str],
+        loop=None,
+        timeout=None,
+        session: aiohttp.ClientSession = None,
+        loads=json.loads, dumps=json.dumps,
+        **kwargs
+    ):
 
-        WSRPCBase.__init__(self, loop=loop, timeout=timeout)
-        self._url = URL(str(endpoint))
-        self._session = session or aiohttp.ClientSession(
-            loop=self._loop, **kwargs
+        WSRPCBase.__init__(
+            self, loop=loop, timeout=timeout, loads=loads, dumps=dumps,
         )
+        self._url = URL(str(endpoint))
+        self._session = session or aiohttp.ClientSession(**kwargs)
+        self.send_lock = Lock()
 
-        self.socket = None      # type: aiohttp.ClientWebSocketResponse
+        self.socket: SocketType = None
         self.closed = False
 
+    # noinspection PyMethodOverriding
     async def close(self):
         """ Close the client connect connection """
 
@@ -47,10 +59,10 @@ class WSRPCClient(WSRPCBase):
 
     async def __handle_connection(self):
         while True:
-            async for message in self.socket:    # type: aiohttp.WSMessage
+            async for message in self.socket:  # type: aiohttp.WSMessage
                 await self._on_message(message)
             else:
-                log.info('Connection was closed')
+                log.info("Connection was closed")
                 self._loop.create_task(self.close())
                 break
 
@@ -58,14 +70,18 @@ class WSRPCClient(WSRPCBase):
         try:
             log.debug(
                 "Sending message to %s serial %s: %s",
-                self._url, Lazy(lambda: kwargs.get('id')),
+                self._url,
+                Lazy(lambda: kwargs.get("id")),
                 Lazy(lambda: kwargs),
             )
 
             if self.socket.closed:
-                raise aiohttp.ClientConnectionError('Connection was closed.')
+                raise aiohttp.ClientConnectionError("Connection was closed.")
 
-            return await self.socket.send_json(kwargs, dumps=lambda x: dumps(x))
+            async with self.send_lock:
+                return await self.socket.send_json(
+                    kwargs, dumps=self._dumps,
+                )
         except aiohttp.WebSocketError:
             self._loop.create_task(self.close())
             raise
@@ -86,4 +102,4 @@ class WSRPCClient(WSRPCBase):
         await self.close()
 
 
-__all__ = 'WSRPCClient',
+__all__ = ("WSRPCClient",)
