@@ -1,112 +1,132 @@
 import abc
-import types
-from collections import defaultdict
-from enum import IntEnum
-from functools import partial, wraps
-
 import asyncio
+import json
 import logging
-from typing import Union, Callable, Any, Dict, Mapping
+import types
+import typing as t
+from collections import defaultdict
+from functools import partial
 
 import aiohttp
 
-from .tools import loads
-from .route import WebSocketRoute
+from wsrpc_aiohttp.signal import Signal
+
+from . import decorators
+from .abc import (
+    AbstractWSRPC, ClientCollectionType, DumpsType, EventListenerCollectionType,
+    EventListenerType, FrameMappingItemType, FutureCollectionType, LoadsType,
+    LocksCollectionType, Proxy, RouteCollectionType, RouteType, TimeoutType,
+)
+from .route import Route
+from .tools import Singleton, awaitable, serializer
 
 
-class ClientException(Exception):
+class WSRPCError(Exception):
     pass
 
 
-class PingTimeoutError(Exception):
+class ClientException(WSRPCError):
+    __slots__ = ("type", "message", "raw")
+
+    def __init__(self, payload):
+        self.type = payload.get("type")
+        self.message = payload.get("message")
+        self.raw = payload
+
+
+class PingTimeoutError(WSRPCError):
     pass
 
 
-def ping(obj, **kwargs):
+def ping(_, **kwargs):
     return kwargs
 
 
 log = logging.getLogger(__name__)
-RouteType = Union[Callable[['WSRPCBase', Any], Any], WebSocketRoute]
-FrameMappingItemType = Mapping[IntEnum, Callable[[aiohttp.WSMessage], Any]]
 
 
-def awaitable(func):
-    if asyncio.iscoroutinefunction(func):
-        return func
-
-    @wraps(func)
-    async def wrap(*args, **kwargs):
-        result = func(*args, **kwargs)
-
-        is_awaitable = (
-            asyncio.iscoroutine(result) or
-            asyncio.isfuture(result) or
-            hasattr(result, '__await__')
-        )
-        if is_awaitable:
-            return await result
-        return result
-
-    return wrap
+class Nothing(Singleton):
+    def __repr__(self):
+        return self.__class__.__name__
 
 
-class _ProxyMethod:
-    __slots__ = '__call', '__name'
-
-    def __init__(self, call_method, name):
-        self.__call = call_method
-        self.__name = name
-
-    def __call__(self, **kwargs):
-        return self.__call(self.__name, **kwargs)
-
-    def __getattr__(self, item: str):
-        return self.__class__(self.__call, ".".join((self.__name, item)))
+CallItem = t.NamedTuple(
+    "CallItem",
+    (
+        ("serial", t.Optional[int]),
+        ("method", t.Union[Nothing, str, None]),
+        ("error", t.Union[Nothing, t.Any]),
+        ("result", t.Union[Nothing, t.Any]),
+        ("params", t.Optional[t.Union[t.List, t.Dict]]),
+    ),
+)
 
 
-class _Proxy:
-    __slots__ = '__call',
-
-    def __init__(self, call_method):
-        self.__call = call_method
-
-    def __getattr__(self, item: str):
-        return _ProxyMethod(self.__call, item)
+def _route_maker() -> t.Dict[str, RouteType]:
+    return {"ping": ping}   # type: ignore
 
 
-class WSRPCBase:
+class WSRPCBase(AbstractWSRPC):
     """ Common WSRPC abstraction """
 
-    _ROUTES = defaultdict(lambda: {'ping': ping})
-    _CLIENTS = defaultdict(dict)
-    _CLEAN_LOCK_TIMEOUT = 2
+    _ROUTES: RouteCollectionType = defaultdict(_route_maker)
+    _CLIENTS: ClientCollectionType = defaultdict(dict)
+    _CLEAN_LOCK_TIMEOUT: t.Union[int, float] = 2
 
-    __slots__ = ('_handlers', '_loop', '_pending_tasks', '_locks',
-                 '_futures', '_serial', '_timeout', '_event_listeners',
-                 '_message_type_mapping')
+    __slots__ = (
+        "_handlers",
+        "_loop",
+        "_pending_tasks",
+        "_locks",
+        "_futures",
+        "_serial",
+        "_timeout",
+        "_event_listeners",
+        "_message_type_mapping",
+    )
 
-    def __init__(self, loop: asyncio.AbstractEventLoop = None, timeout=None):
+    ON_CALL_START = Signal()
+    ON_CALL_SUCCESS = Signal()
+    ON_CALL_FAIL = Signal()
+
+    _pending_tasks: t.Set[t.Union[asyncio.Task, asyncio.Handle]]
+    _handlers: t.Dict[str, RouteType]
+
+    def _dumps(self, value: t.Any) -> t.Any:
+        return self._json_dumps(value, default=serializer)
+
+    def __init__(
+        self, loop: asyncio.AbstractEventLoop = None,
+        timeout: t.Optional[TimeoutType] = None,
+        loads: LoadsType = json.loads,
+        dumps: DumpsType = json.dumps,
+    ):
+        self._json_dumps = dumps
+        self._json_loads = loads
         self._loop = loop or asyncio.get_event_loop()
         self._handlers = {}
         self._pending_tasks = set()
         self._serial = 0
-        self._timeout = timeout
-        self._locks = defaultdict(asyncio.Lock)
-        self._futures = defaultdict(self._loop.create_future)
-        self._event_listeners = set()
+        self._timeout: t.Optional[TimeoutType] = timeout
+        self._locks: LocksCollectionType = defaultdict(asyncio.Lock)
+        self._futures: FutureCollectionType = defaultdict(
+            self._loop.create_future,
+        )
+        self._event_listeners: EventListenerCollectionType = set()
         self._message_type_mapping = self._create_type_mapping()
 
     def _create_type_mapping(self) -> FrameMappingItemType:
-        return types.MappingProxyType({
-            aiohttp.WSMsgType.TEXT: self.handle_message,
-            aiohttp.WSMsgType.BINARY: self.handle_binary,
-            aiohttp.WSMsgType.CLOSE: self.close,
-            aiohttp.WSMsgType.CLOSED: self.close,
-        })
+        return types.MappingProxyType(
+            {
+                aiohttp.WSMsgType.TEXT: self.handle_message,
+                aiohttp.WSMsgType.BINARY: self.handle_binary,
+                aiohttp.WSMsgType.CLOSE: self.close,
+                aiohttp.WSMsgType.CLOSED: self.close,
+            },
+        )
 
     def _create_task(self, coro):
-        task = self._loop.create_task(coro)      # type: asyncio.Task
+        task: asyncio.Task = self._loop.create_task(coro)
         self._pending_tasks.add(task)
         task.add_done_callback(partial(self._pending_tasks.remove))
 
@@ -120,8 +140,9 @@ class WSRPCBase:
 
     async def close(self, message=None):
         """ Cancel all pending tasks """
+
         async def task_waiter(task):
-            if not (hasattr(task, '__iter__') or hasattr(task, '__aiter__')):
+            if not (hasattr(task, "__iter__") or hasattr(task, "__aiter__")):
                 return
 
             try:
@@ -130,7 +151,7 @@ class WSRPCBase:
                 pass
             except Exception:
                 log.exception(
-                    "Unhandled exception when closing client connection"
+                    "Unhandled exception when closing client connection",
                 )
 
         if message:
@@ -139,53 +160,87 @@ class WSRPCBase:
         for task in tuple(self._pending_tasks):
             task.cancel()
 
-            if hasattr(task, 'cancelled') and not task.cancelled():
+            if hasattr(task, "cancelled") and not task.cancelled():
                 self._loop.create_task(task_waiter(task))
 
     async def handle_binary(self, message: aiohttp.WSMessage):
         log.warning("Unhandled message %r %r", message.type, message.data)
 
-    async def handle_message(self, message: aiohttp.WSMessage):
-        # deserialize message
-        # noinspection PyNoneFunctionAssignment, PyTypeChecker
-        data = message.json(loads=loads)  # type: dict
+    async def _call_method(self, call_item: CallItem):
+        try:
+            if not isinstance(call_item.method, Nothing) and call_item.serial:
+                log.debug(
+                    "Acquiring lock for %r serial %r", self, call_item.serial,
+                )
+                async with self._locks[call_item.serial]:
+                    args, kwargs = self.prepare_args(call_item.params)
 
-        log.debug("Response: %r", data)
-        serial = data.get('id')
+                    return await self.handle_method(
+                        call_item.method, call_item.serial, args, kwargs,
+                    )
+            elif not isinstance(call_item.result, Nothing):
+                return await self.handle_result(
+                    call_item.serial, call_item.result,
+                )
+            elif not isinstance(call_item.error, Nothing):
+                return await self.handle_error(
+                    call_item.serial, call_item.error,
+                )
+            else:
+                return await self.handle_result(call_item.serial, None)
+
+        except Exception as e:
+            log.exception(e)
+
+            if call_item.serial:
+                await self._send(
+                    error=self._format_error(e), id=call_item.serial,
+                )
+        finally:
+            self._call_later(
+                self._CLEAN_LOCK_TIMEOUT, self.__clean_lock, call_item.serial,
+            )
+
+    @staticmethod
+    def _parse_message(data: dict) -> CallItem:
+        message_id = data.get("id")     # type: t.Optional[int]
+
+        if message_id and not isinstance(message_id, int):
+            raise ValueError
+
+        message_method: t.Union[str, Nothing, None] = data.get(
+            "method", Nothing(),
+        )
+
+        message_result: t.Union[str, Nothing, None] = data.get(
+            "result", Nothing(),
+        )
+
+        message_error: t.Union[str, Nothing, None] = data.get(
+            "error", Nothing(),
+        )
+
+        message_params: t.Union[t.List[t.Any], t.Dict[t.Any, t.Any], None] = (
+            data.get("params", None)
+        )
+
+        return CallItem(
+            serial=message_id, method=message_method, result=message_result,
+            error=message_error, params=message_params,
+        )
+
+    async def handle_message(self, message: aiohttp.WSMessage):
+        # noinspection PyTypeChecker, PyNoneFunctionAssignment
+        data: dict = message.json(loads=self._json_loads)
+        log.debug("Got message: %r", data)
+
+        serial = data.get("id")
 
         if serial is None:
             return await self.handle_event(data)
 
-        method = data.get('method')
-        result = data.get('result')
-        error = data.get('error')
-
-        log.debug("Acquiring lock for %s serial %s", self, serial)
-        async with self._locks[serial]:
-            try:
-                if 'method' in data:
-                    args, kwargs = self.prepare_args(
-                        data.get('params', None)
-                    )
-                    return await self.handle_method(
-                        method, serial, *args, **kwargs
-                    )
-                elif 'result' in data:
-                    return await self.handle_result(serial, result)
-                elif 'error' in data:
-                    return await self.handle_error(serial, error)
-                else:
-                    return await self.handle_result(serial, None)
-
-            except Exception as e:
-                log.exception(e)
-
-                if serial:
-                    await self._send(error=self._format_error(e), id=serial)
-            finally:
-                self._call_later(
-                    self._CLEAN_LOCK_TIMEOUT, self.__clean_lock, serial
-                )
+        call_item = self._parse_message(data)
+        await self._call_method(call_item)
 
     async def _on_message(self, msg: aiohttp.WSMessage):
         async def unknown_method(msg: aiohttp.WSMessage):
@@ -195,20 +250,20 @@ class WSRPCBase:
         self._create_task(awaitable(handler)(msg))
 
     @classmethod
-    def get_routes(cls) -> Dict[str, RouteType]:
+    def get_routes(cls) -> t.Dict[str, RouteType]:
         return cls._ROUTES[cls]
 
     @classmethod
-    def get_clients(cls) -> Dict[str, 'WSRPCBase']:
+    def get_clients(cls) -> t.Dict[str, AbstractWSRPC]:
         return cls._CLIENTS[cls]
 
     @property
-    def routes(self) -> Dict[str, RouteType]:
+    def routes(self) -> t.Dict[str, RouteType]:
         """ Property which contains the socket routes """
         return self.get_routes()
 
     @property
-    def clients(self) -> Dict[str, 'WSRPCBase']:
+    def clients(self) -> t.Dict[str, AbstractWSRPC]:
         """ Property which contains the socket clients """
         return self.get_clients()
 
@@ -234,12 +289,15 @@ class WSRPCBase:
 
     @staticmethod
     def is_route(func):
-        return (
-            hasattr(func, '__self__') and
-            isinstance(func.__self__, WebSocketRoute)
-        )
+        return hasattr(func, "__self__") and isinstance(func.__self__, Route)
 
-    async def handle_method(self, method, serial, *args, **kwargs):
+    async def handle_method(self, method, serial, args, kwargs):
+        await self.ON_CALL_START.call(
+            method=method,
+            serial=serial,
+            args=args,
+            kwargs=kwargs,
+        )
         callee = self.resolver(method)
 
         if not self.is_route(callee):
@@ -247,7 +305,27 @@ class WSRPCBase:
             a.extend(args)
             args = a
 
-        result = await self._executor(partial(callee, *args, **kwargs))
+        func = partial(callee, *args, **kwargs)
+        try:
+            result = await self._executor(func)
+        except Exception as err:
+            await self.ON_CALL_FAIL.call(
+                method=method,
+                serial=serial,
+                args=args,
+                kwargs=kwargs,
+                err=err,
+            )
+            raise
+
+        await self.ON_CALL_SUCCESS.call(
+            method=method,
+            serial=serial,
+            args=args,
+            kwargs=kwargs,
+            result=result,
+        )
+
         await self._send(result=result, id=serial)
 
     async def handle_result(self, serial, result):
@@ -258,7 +336,7 @@ class WSRPCBase:
 
     async def handle_error(self, serial, error):
         self._reject(serial, error)
-        log.error('Client return error: \n\t%r', error)
+        log.error("Client return error: \n\t%r", error)
 
     def __clean_lock(self, serial):
         if serial not in self._locks:
@@ -276,7 +354,7 @@ class WSRPCBase:
 
     @staticmethod
     def _format_error(e):
-        return {'type': str(type(e).__name__), 'message': str(e)}
+        return {"type": str(type(e).__name__), "message": str(e)}
 
     def _reject(self, serial, error):
         future = self._futures.get(serial)
@@ -288,21 +366,27 @@ class WSRPCBase:
 
     def _unresolvable(self, func_name, *args, **kwargs):
         raise NotImplementedError(
-            'Callback function "%r" not implemented' % func_name
+            'Callback function "%r" not implemented' % func_name,
         )
 
     def resolver(self, func_name):
         class_name, method = (
-            func_name.split('.') if '.' in func_name else (func_name, 'init')
+            func_name.split(".", 1)
+            if "." in func_name
+            else (func_name, "init")
         )
 
         callee = self.routes.get(class_name, self._unresolvable)
 
+        if isinstance(callee, decorators.ProxyBase):
+            callee = callee.func
+
         condition = (
-            callee == self._unresolvable or
-            isinstance(getattr(callee, '__self__', None), WebSocketRoute) or (
+            callee == self._unresolvable
+            or isinstance(getattr(callee, "__self__", None), Route)
+            or (
                 not isinstance(callee, (types.FunctionType, types.MethodType))
-                and issubclass(callee, WebSocketRoute)
+                and issubclass(callee, Route)
             )
         )
 
@@ -310,21 +394,21 @@ class WSRPCBase:
             if class_name not in self._handlers:
                 self._handlers[class_name] = callee(self)
 
-            return self._handlers[class_name]._resolve(method)  # noqa
+            return self._handlers[class_name](method)
 
         callee = self.routes.get(func_name, self._unresolvable)
-        if hasattr(callee, '__call__'):
+        if hasattr(callee, "__call__"):
             return callee
         else:
             raise NotImplementedError(
-                'Method call of {0} is not implemented'.format(repr(callee))
+                "Method call of {0} is not implemented".format(repr(callee)),
             )
 
     def _get_serial(self):
         self._serial += 2
         return self._serial
 
-    async def call(self, func: str, **kwargs):
+    async def call(self, func: str, timeout=None, **kwargs):
         """ Method for call remote function
 
         Remote methods allows only kwargs as arguments.
@@ -356,18 +440,23 @@ class WSRPCBase:
         payload = dict(id=serial, method=func, params=kwargs)
 
         log.info(
-            "Sending request #%r \"%s(%r)\" to the client.",
-            serial, func, kwargs,
+            'Sending request #%r "%s(%r)" to the client.',
+            serial,
+            func,
+            kwargs,
         )
 
         await self._send(**payload)
-        return await asyncio.wait_for(future, self._timeout)
+        result = await asyncio.wait_for(
+            future, timeout=timeout or self._timeout,
+        )
+        return result
 
     async def emit(self, event):
         await self._send(**event)
 
     @classmethod
-    def add_route(cls, route: str, handler: Union[WebSocketRoute, Callable]):
+    def add_route(cls, route: str, handler: RouteType) -> None:
         """ Expose local function through RPC
 
         :param route: Name which function will be aliased for this function.
@@ -387,10 +476,13 @@ class WSRPCBase:
             without params before callable method.
 
         """
-        assert callable(handler) or isinstance(handler, WebSocketRoute)
+        assert callable(handler) or isinstance(handler, Route)
+        if callable(handler):
+            handler = decorators.proxy(handler)
+
         cls.get_routes()[route] = handler
 
-    def add_event_listener(self, func: Callable[[dict], Any]):
+    def add_event_listener(self, func: EventListenerType):
         self._event_listeners.add(func)
 
     def remove_event_listeners(self, func):
@@ -407,7 +499,7 @@ class WSRPCBase:
             cls.get_routes().pop(route, None)
 
     def __repr__(self):
-        if hasattr(self, 'id'):
+        if hasattr(self, "id"):
             return "<RPCWebSocket: ID[{0}]>".format(self.id)
         else:
             return "<RPCWebsocket: {0} (waiting)>".format(self.__hash__())
@@ -429,4 +521,7 @@ class WSRPCBase:
             # full equivalent of
             await client.call('ping')
         """
-        return _Proxy(self.call)
+        return Proxy(self.call)
+
+
+__all__ = ("Route", "WSRPCBase", "ClientException", "WSRPCError")
