@@ -1,16 +1,22 @@
 # encoding: utf-8
 import asyncio
+import json
 import logging
 import uuid
 from collections import defaultdict
+from functools import partial
+from typing import Optional
 
 import aiohttp
-from aiohttp import web, WebSocketError
+from aiohttp import WebSocketError, web
 from aiohttp.abc import AbstractView
 
-from .common import WSRPCBase, ClientException, awaitable
-from .route import WebSocketRoute
-from .tools import Lazy, dumps
+from wsrpc_aiohttp.signal import Signal
+
+from .abc import TimeoutType
+from .common import ClientException, WSRPCBase
+from .tools import Lazy, awaitable
+
 
 global_log = logging.getLogger("wsrpc")
 log = logging.getLogger("wsrpc.handler")
@@ -19,17 +25,38 @@ log = logging.getLogger("wsrpc.handler")
 class WebSocketBase(WSRPCBase, AbstractView):
     """ Base class for aiohttp websocket handler """
 
-    __slots__ = ('_request', 'socket', 'id', '__pending_tasks',
-                 '__handlers', 'store', 'serial', '_ping', 'protocol_version')
+    __slots__ = (
+        "_request",
+        "socket",
+        "id",
+        "__pending_tasks",
+        "__handlers",
+        "store",
+        "serial",
+        "_ping",
+        "protocol_version",
+    )
 
-    KEEPALIVE_PING_TIMEOUT = 30        # type: int
-    CLIENT_TIMEOUT = int(KEEPALIVE_PING_TIMEOUT / 3)    # type: int
-    MAX_CONCURRENT_REQUESTS = 25       # type: int
-    REQUEST_EXECUTION_TIMEOUT = None   # type: int
+    KEEPALIVE_PING_TIMEOUT: TimeoutType = 30
+    CLIENT_TIMEOUT: TimeoutType = int(KEEPALIVE_PING_TIMEOUT / 3)
+    MAX_CONCURRENT_REQUESTS: TimeoutType = 25
+    REQUEST_EXECUTION_TIMEOUT: Optional[TimeoutType] = None
+
+    JSON_LOADS = staticmethod(json.loads)
+    JSON_DUMPS = staticmethod(json.dumps)
+
+    ON_AUTH_SUCCESS = Signal()
+    ON_AUTH_FAIL = Signal()
+    ON_CONN_OPEN = Signal()
+    ON_CONN_CLOSE = Signal()
+    ON_CONN_FAIL = Signal()
 
     def __init__(self, request):
         AbstractView.__init__(self, request)
-        WSRPCBase.__init__(self, timeout=self.REQUEST_EXECUTION_TIMEOUT)
+        WSRPCBase.__init__(
+            self, timeout=self.REQUEST_EXECUTION_TIMEOUT,
+            loads=self.JSON_LOADS, dumps=self.JSON_DUMPS,
+        )
 
         self._ping = defaultdict(self._loop.create_future)
         self.id = uuid.uuid4()
@@ -38,11 +65,18 @@ class WebSocketBase(WSRPCBase, AbstractView):
         self.semaphore = asyncio.Semaphore(self.MAX_CONCURRENT_REQUESTS)
 
     @classmethod
-    def configure(cls, keepalive_timeout=KEEPALIVE_PING_TIMEOUT,
-                  client_timeout=CLIENT_TIMEOUT,
-                  max_concurrent_requests=MAX_CONCURRENT_REQUESTS):
+    def configure(
+        cls,
+        keepalive_timeout=KEEPALIVE_PING_TIMEOUT,
+        client_timeout=CLIENT_TIMEOUT,
+        max_concurrent_requests=MAX_CONCURRENT_REQUESTS,
+        loads=json.loads,
+        dumps=json.dumps,
+    ):
         """ Configures the handler class
 
+        :param dumps: json serializer
+        :param loads: json deserializer
         :param keepalive_timeout: sets timeout of client pong response
         :param client_timeout: internal lock timeout
         :param max_concurrent_requests: how many concurrent requests might
@@ -52,6 +86,24 @@ class WebSocketBase(WSRPCBase, AbstractView):
         cls.KEEPALIVE_PING_TIMEOUT = keepalive_timeout
         cls.CLIENT_TIMEOUT = client_timeout
         cls.MAX_CONCURRENT_REQUESTS = max_concurrent_requests
+        cls.JSON_LOADS = staticmethod(loads)
+        cls.JSON_DUMPS = staticmethod(dumps)
+
+    @classmethod
+    def freeze(cls):
+        """ Freeze all signals """
+        for signal in (
+            cls.ON_AUTH_SUCCESS,
+            cls.ON_AUTH_FAIL,
+            cls.ON_CONN_OPEN,
+            cls.ON_CONN_CLOSE,
+            cls.ON_CONN_FAIL,
+            cls.ON_CALL_START,
+            cls.ON_CALL_SUCCESS,
+            cls.ON_CALL_FAIL,
+        ):  # type: Signal
+            if not signal.is_frozen:
+                signal.freeze()
 
     def __await__(self):
         return self.__handle_request().__await__()
@@ -75,10 +127,29 @@ class WebSocketBase(WSRPCBase, AbstractView):
     async def __handle_request(self):
         self.socket = web.WebSocketResponse()
 
+        await self.ON_CONN_OPEN.call(socket=self.socket, request=self.request)
+
         if not await self.authorize():
+            await self.ON_AUTH_FAIL.call(
+                socket=self.socket,
+                request=self.request,
+            )
             raise web.HTTPForbidden()
 
-        await self.socket.prepare(self.request)
+        await self.ON_AUTH_SUCCESS.call(
+            socket=self.socket,
+            request=self.request,
+        )
+
+        try:
+            await self.socket.prepare(self.request)
+        except Exception as err:
+            await self.ON_CONN_FAIL.call(
+                socket=self.socket,
+                request=self.request,
+                err=err,
+            )
+            raise
 
         try:
             self.clients[self.id] = self
@@ -88,57 +159,67 @@ class WebSocketBase(WSRPCBase, AbstractView):
                 try:
                     await self._on_message(msg)
                 except WebSocketError:
-                    log.error('Client connection %s closed with exception %s',
-                              self.id, self.socket.exception())
+                    log.error(
+                        "Client connection %s closed with exception %s",
+                        self.id,
+                        self.socket.exception(),
+                    )
                     break
             else:
-                log.info('Client connection %s closed', self.id)
+                log.info("Client connection %s closed", self.id)
 
             return self.socket
         finally:
+            await self.ON_CONN_CLOSE.call(
+                socket=self.socket,
+                request=self.request,
+            )
             await self.close()
 
     @classmethod
-    def broadcast(cls, func, callback=WebSocketRoute.placebo, **kwargs):
+    def broadcast(cls, func, callback=None, return_exceptions=True, **kwargs):
         """ Call remote function on all connected clients
 
         :param func: Remote route name
         :param callback: Function which receive responses
+        :param return_exceptions: Return exceptions of client calls
+            instead of raise a first one
         """
 
-        loop = asyncio.get_event_loop()
-
         tasks = []
 
         for client in cls.get_clients().values():
-            tasks.append(loop.create_task(
-                client.call, func, callback, **kwargs
-            ))
+            task = asyncio.ensure_future(client.call(func, **kwargs))
 
-        return asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
+            if callback:
+                task.add_done_callback(partial(callback, client))
+
+            tasks.append(task)
+
+        return asyncio.gather(*tasks, return_exceptions=return_exceptions)
 
     async def _send(self, **kwargs):
         try:
             log.debug(
                 "Sending message to %s serial %s: %s",
                 Lazy(lambda: str(self.id)),
-                Lazy(lambda: str(kwargs.get('id'))),
-                Lazy(lambda: str(kwargs))
-              )
-            await self.socket.send_json(kwargs, dumps=lambda x: dumps(x))
+                Lazy(lambda: str(kwargs.get("id"))),
+                Lazy(lambda: str(kwargs)),
+            )
+            await self.socket.send_json(kwargs, dumps=self._json_dumps)
         except aiohttp.WebSocketError:
             self._create_task(self.close())
 
     @staticmethod
     def _format_error(e):
-        return {'type': str(type(e).__name__), 'message': str(e)}
+        return {"type": str(type(e).__name__), "message": str(e)}
 
     def _reject(self, serial, error):
         future = self._futures.get(serial)
         if future:
             future.set_exception(ClientException(error))
 
-    async def close(self):
+    async def close(self, message=None):
         """ Cancel all pending tasks and stop this socket connection """
         await self.socket.close()
         await super().close()
@@ -150,9 +231,12 @@ class WebSocketBase(WSRPCBase, AbstractView):
             self._loop.create_task(awaitable(obj._onclose)())
 
     def _log_client_list(self):
-        log.debug('CLIENTS: %s', Lazy(lambda: ''.join([
-            '\n\t%r' % i for i in self.clients.values()
-        ])))
+        log.debug(
+            "CLIENTS: %s",
+            Lazy(
+                lambda: "".join(["\n\t%r" % i for i in self.clients.values()]),
+            ),
+        )
 
     async def _start_ping(self):
         while True:
@@ -160,7 +244,7 @@ class WebSocketBase(WSRPCBase, AbstractView):
                 return
 
             future = asyncio.ensure_future(
-                self.call('ping', seq=self._loop.time())
+                self.call("ping", seq=self._loop.time()),
             )
 
             def on_timeout():
@@ -174,7 +258,7 @@ class WebSocketBase(WSRPCBase, AbstractView):
                 future.set_exception(TimeoutError)
 
             handle = self._loop.call_later(
-                self.KEEPALIVE_PING_TIMEOUT, on_timeout
+                self.KEEPALIVE_PING_TIMEOUT, on_timeout,
             )
 
             future.add_done_callback(lambda f: handle.cancel())
@@ -184,25 +268,31 @@ class WebSocketBase(WSRPCBase, AbstractView):
                 if not resp:
                     continue
 
-                delta = (self._loop.time() - resp.get('seq', 0))
+                delta = self._loop.time() - resp.get("seq", 0)
 
                 log.debug("%r Pong recieved: %.4f" % (self, delta))
 
             except asyncio.CancelledError:
                 break
             except (TimeoutError, asyncio.TimeoutError):
-                log.info('Client "%r" connection should be '
-                         'closed because ping timeout', self)
+                log.info(
+                    'Client "%r" connection should be '
+                    "closed because ping timeout",
+                    self,
+                )
 
                 self._loop.create_task(self.close())
                 break
             except Exception:
-                log.exception('Error when ping remote side.')
+                log.exception("Error when ping remote side.")
                 break
 
             if delta > self.CLIENT_TIMEOUT:
-                log.info('Client "%r" connection should be closed because ping '
-                         'response time gather then client timeout', self)
+                log.info(
+                    'Client "%r" connection should be closed because ping '
+                    "response time gather then client timeout",
+                    self,
+                )
                 self._loop.create_task(self.close())
                 break
 
@@ -211,6 +301,7 @@ class WebSocketBase(WSRPCBase, AbstractView):
 
 class WebSocketAsync(WebSocketBase):
     """ Handler class which execute any route as a coroutine """
+
     async def _executor(self, func):
         return await awaitable(func)()
 
@@ -223,4 +314,9 @@ class WebSocketThreaded(WebSocketBase):
         return await self._loop.run_in_executor(None, func)
 
 
-__all__ = ('WebSocketAsync', 'WebSocketThreaded', 'WebSocketBase')
+__all__ = (
+    "ClientException",
+    "WebSocketAsync",
+    "WebSocketBase",
+    "WebSocketThreaded",
+)
